# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2013 IBM Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyghmi.constants as const

IPMI_BMC_ADDRESS = 0x20
IPMI_SEND_MESSAGE_CMD = 0x34


payload_types = {
    'ipmi': 0x0,
    'sol': 0x1,
    'rmcpplusopenreq': 0x10,
    'rmcpplusopenresponse': 0x11,
    'rakp1': 0x12,
    'rakp2': 0x13,
    'rakp3': 0x14,
    'rakp4': 0x15,
}

# sensor type codes, table 42-3
sensor_type_codes = {
    1: 'Temperature',
    2: 'Voltage',
    3: 'Current',
    4: 'Fan',
    5: 'Physical Security',
    6: 'Platform Security',
    7: 'Processor',
    8: 'Power Supply',
    9: 'Power Unit',
    0xa: 'Cooling Device',
    0xb: 'Other',
    0xc: 'Memory',
    0xd: 'Drive Bay',
    0xe: 'POST Memory Resize',
    0xf: 'System Firmware',
    0x10: 'Event Log Disabled',
    0x11: 'Watchdog',
    0x12: 'System Event',
    0x13: 'Critical interrupt',
    0x14: 'Button/switch',
    0x15: 'Module/Board',
    0x16: 'Microcontroller/Coprocessor',
    0x17: 'Add-in Card',
    0x18: 'Chassis',
    0x19: 'Chip Set',
    0x1a: 'Other FRU',
    0x1b: 'Cable/Interconnect',
    0x1c: 'Terminator',
    0x1d: 'System Boot',
    0x1e: 'Boot Error',
    0x1f: 'OS Boot',
    0x20: 'OS Stop',
    0x21: 'Slot/Connector',
    0x22: 'System ACPI Power State',
    0x23: 'Watchdog',
    0x24: 'Platform alert',
    0x25: 'Entity Presence',
    0x26: 'Monitor ASIC/IC',
    0x27: 'LAN',
    0x28: 'Management Subsystem Health',
    0x29: 'Battery',
    0x2a: 'Session',
    0x2b: 'Version Change',
    0x2c: 'FRU State',
}

# This is from table 42-2
# digital discrete poses a challenge from a health perspective.  So far all
# observed ones are no more or less 'healthy' by being asserted or not asserted
# for example asserting that an add-on is installed

generic_type_offsets = {
    1: {  # threshold based
        # Some explanation is offered in the specification around 'get sensor
        # event status' command.  Assertions should indicate the new state
        # and deassertion should indicate leaving the state.  The 'going high'
        # and 'going low' do not denote leaving a state.  For example, going
        # from Lower Critical to Lower Non-Critical would be 'going high'
        # Will just report the new state rather than the direction from
        # which things came from, since in a vacuum it is not useful data
        # and in the context of event log, it should be better discerned
        # from the pattern of prior events
        0: {
            'desc': 'Lower non-critical',  # - going low',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Lower Non-critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Lower non-critical',  # - going high',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Lower non-critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Lower Critical',  # - going low',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Lower critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Lower Critical',  # - going high',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Lower critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Lower non-recoverable',  # - going low
            'severity': const.Health.Failed,
            'deassertion_desc': 'Lower non-recoverable deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Lower non-recoverable',  # - going high
            'severity': const.Health.Failed,
            'deassertion_desc': 'Lower non-recoverable deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Upper non-critical',  # - going low
            'severity': const.Health.Warning,
            'deassertion_desc': 'Upper non-critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Upper non-critical',  # - going high
            'severity': const.Health.Warning,
            'deassertion_desc': 'Upper non-critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Upper critical',  # - going low
            'severity': const.Health.Critical,
            'deassertion_desc': 'Upper critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        9: {
            'desc': 'Upper critical',  # - going high
            'severity': const.Health.Critical,
            'deassertion_desc': 'Upper critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xa: {
            'desc': 'Upper non-recoverable',  # - going low
            'severity': const.Health.Failed,
            'deassertion_desc': 'Upper non-recoverable deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xb: {
            'desc': 'Upper non-recoverable',  # - going high
            'severity': const.Health.Failed,
            'deassertion_desc': 'Upper non-recoverable deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    2: {
        0: {
            'desc': 'Idle',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Idle deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Active',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Active deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Busy',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Busy deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    3: {
        0: {
            'desc': 'Deasserted',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Asserted',  # this would be very peculiar
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Asserted',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    4: {
        0: {
            'desc': 'Predictive failure deasserted',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Predictive failure',
            'deassertion_severity': const.Health.Warning,
        },
        1: {
            'desc': 'Predictive failure',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Predictive failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    5: {
        0: {
            'desc': 'Limit not exceeded',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Limit exceeded',
            'deassertion_severity': const.Health.Warning,
        },
        1: {
            'desc': 'Limit exceeded',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Limit no longer exceeded',
            'deassertion_severity': const.Health.Ok,
        },
    },
    6: {
        0: {
            'desc': 'Performance met',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Performance lags',
            'deassertion_severity': const.Health.Warning,
        },
        1: {
            'desc': 'Performance lags',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Performance no longer lags',
            'deassertion_severity': const.Health.Ok,
        },
    },
    7: {
        0: {
            'desc': 'Ok',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Not ok',
            'deassertion_severity': const.Health.Warning,
        },
        1: {
            'desc': 'Non-Critical',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Non-critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Critical',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Non-recoverable',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Non-recoverable deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Non-critical',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Non-critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Critical',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Critical deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Non-recoverable',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Non-recoverable deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Monitor',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Monitor deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Informational',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Informational deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    8: {
        0: {
            'desc': 'Absent',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Present',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
    },
    9: {
        0: {
            'desc': 'Disabled',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Enabled',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Enabled',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Disabled',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0xa: {
        0: {
            'desc': 'Running',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Running deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'In test',
            'severity': const.Health.Ok,
            'deassertion_desc': 'In test deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Power off',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Power off deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Online',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Online deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Offline',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Offline deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Off duty',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Off duty deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Degraded',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Degraded deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Power save',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Power save deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Install error',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Install error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0xb: {
        0: {
            'desc': 'Redundant',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Redundant lost',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Not redundant',  # redundancy lost
            'severity': const.Health.Warning,
            'deassertion_desc': 'Redundancy restored',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Redundancy degraded',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Redundancy restored',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Not redundant',  # down to non redundant
            'severity': const.Health.Ok,
            'deassertion_desc': 'Redundancy restored',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Not Redundant',  # it has come back, but not redundant
            'severity': const.Health.Warning,
            'deassertion_desc': 'Redundancy restored',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Degraded',  # abnormal operation
            'severity': const.Health.Critical,
            'deassertion_desc': 'Present',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Redundancy degraded',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Redundancy restored',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Redundancy degraded',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Redundancy restored',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0xc: {
        0: {
            'desc': 'ACPI D0',
            'severity': const.Health.Ok,
            'deassertion_desc': 'ACPI D0 deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'ACPI D1',
            'severity': const.Health.Ok,
            'deassertion_desc': 'ACPI D1 deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'ACPI D2',
            'severity': const.Health.Ok,
            'deassertion_desc': 'ACPI D2 deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'ACPI D3',
            'severity': const.Health.Ok,
            'deassertion_desc': 'ACPI D3 deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
}

sensor_type_offsets = {
    # For the security sensors, we assume if armed,
    # the operator considers these to  be critical situations
    # Even though many of these would never reasonably be 'deasserted', provide
    # some language that hypothetically would make sense if it were ever
    # observed
    5: {
        0: {
            'desc': 'Chassis intrusion',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Chassis intrusion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Drive area intrusion',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Drive area intrusion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'I/O card area intrusion',
            'severity': const.Health.Critical,
            'deassertion_desc': 'I/O card area intrusion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Processor area intrusion',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Processor area intrusion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'LAN connection lost',
            'severity': const.Health.Critical,
            'deassertion_desc': 'LAN connection restored',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Unauthorized dock',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Unauthorized dock deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Fan area intrusion',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Fan area intrusion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    6: {
        0: {
            'desc': 'Front Panel Lockout Violation attempt',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Front Panel Lockout deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Pre-boot password violation - user',
            'severity': const.Health.Critical,
            'deassertion_severity': const.Health.Ok,
            'deassertion_desc': 'Boot Password violation deasserted',
        },
        2: {
            'desc': 'Pre-boot password violation - setup',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Setup password violation deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Pre-boot password violation - netboot',
            'severity': const.Health.Critical,
            'deassertion_desc':
            'Pre-boot password violation - netboot deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Pre-boot password violation',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Pre-boot password violation deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Out-of-band access password violation',
            'severity': const.Health.Critical,
            'deassertion_desc':
            'Out-of-band access password violation deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    7: {
        0: {
            'desc': 'Internal error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Internal error recovered',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Thermal trip',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Thermal trip recovered',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Self test failure',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Self test recovered',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Hang in POST',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Hang in POST recovered',
            'deassertion_severity': const.Health.Ok,

        },
        4: {
            'desc': 'Startup failure',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Startup recovered',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Configuration error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Configuration error recovered',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Uncorrectable cpu complex error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Uncorrectable cpu complex error recovered',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Disabled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Enabled',
            'deassertion_severity': const.Health.Ok,
        },
        9: {
            'desc': 'Terminator presence detected',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Terminator absent',
            'deassertion_severity': const.Health.Ok,
        },
        0xa: {
            'desc': 'Throttled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'No longer throttled',
            'deassertion_severity': const.Health.Ok,
        },
        0xb: {
            'desc': 'Uncorrectable machine check exception',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Uncorrectable machine check recovered',
            'deassertion_severity': const.Health.Ok,
        },
        0xc: {
            'desc': 'Correctable machine check exception',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Correctable machine check recovered',
            'deassertion_severity': const.Health.Ok,
        },
    },
    8: {  # power supply
        0: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Failure',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Recovered',
            'deassertion_severity': const.Health.Ok,

        },
        2: {
            'desc': 'Predictive failure',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Predictive failure recovered',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Power input lost',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Power input restored',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Power input out of range or lost',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Power input present and in range',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Power input out of range',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Power input in range',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            # clarified by SEL/PET event data 3
            'desc': 'Configuration error',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Configuration error recovered',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Standby',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Active',
            'deassertion_severity': const.Health.Ok,
        },
    },
    9: {  # power unit
        0: {
            'desc': 'Power off',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Power on',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Power cycle',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Power cycle deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': '240VA power down',
            'severity': const.Health.Warning,
            'deassertion_desc': '240VA power down deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Interlock power down',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Interlock power down deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Power input lost',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Power input restored',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Soft power control failure',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Soft power control failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Power unit failure',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Power unit failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Power unit predictive failure',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Power unit predictive failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0xc: {  # memory
        0: {
            'desc': 'Correctable error',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Correctable error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Uncorrectable error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Uncorrectable error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Parity error',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Parity error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Scrub failed',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Scrub failed deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Device disabled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Device enabled',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Correctable memory error logging limit reached',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Correctable memory error logging restored',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Configuration error',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Configuration error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Spare',  # event data 3 available
            'severity': const.Health.Ok,
            'deassertion_desc': 'Spare deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        9: {
            'desc': 'Throttled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'No longer throttled',
            'deassertion_severity': const.Health.Ok,
        },
        0xa: {
            'desc': 'Critical over temperature',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Critical over temperature deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0xd: {  # drive bay
        0: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Drive fault',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Drive fault deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Predictive drive failure',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Predictive drive failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Hot spare drive',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Hot spare drive deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Drive consistency check in progress',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Drive consistency check complete',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Drive in critical array',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Drive in critical array deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Drive in failed array',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Drive in failed array deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Rebuild in progress',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Rebuild complete',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Rebuild aborted',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Rebuild abort deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0xf: {
        0: {
            'desc': 'Boot error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Boot error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Hang',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Hang deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Progress',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Progress deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x10: {  # event log disabled
        0: {
            'desc': 'Correctable memory error logging disabled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Correctable memory error logging enabled',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Specific event logging disabled',
            'deassertion_desc': 'Event logging enabled',
            'severity': const.Health.Warning,
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Log clear',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Log clear complete',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Logging disabled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Logging enabled',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Event log full',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Event log no longer full',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Event log nearly full',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Event log no longer nearly full',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Correctable machine check logging disabled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Correctable machine check logging enabled',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x12: {  # system event
        0: {
            'desc': 'System reconfigured',
            'severity': const.Health.Ok,
            'deassertion_desc': 'System reconfiguration deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'OEM boot event',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OEM boot event deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Undetermined hardware failure',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Undetermined hardware failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Aux log entry',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Aux log entry deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Event response',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Event response deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Clock time change',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Clock time change deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x13: {  # critical interrupt
        0: {
            'desc': 'Front panel diagnostic interrupt',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Front panel diagnostic interrupt deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Bus timeout',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Bus timeout recovered',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'I/O NMI',
            'severity': const.Health.Critical,
            'deassertion_desc': 'I/O NMI deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Software NMI',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Software NMI deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'PCI PERR',
            'severity': const.Health.Failed,
            'deassertion_desc': 'PCI PERR deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'PCI SERR',
            'severity': const.Health.Failed,
            'deassertion_desc': 'PCI SERR deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'EISA fail safe timeout',
            'severity': const.Health.Failed,
            'deassertion_desc': 'EISA fail safe timeout deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Bus correctable error',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Bus correctable error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'Bus uncorrectable error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Bus uncorrectable error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        9: {
            'desc': 'Fatal NMI',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Fatal NMI deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xa: {
            'desc': 'Bus fatal error',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Bus fatal error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xb: {
            'desc': 'Bus degraded',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Bus degraded deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x14: {  # button/switch
        0: {
            'desc': 'Power button pressed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Power button released',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Sleep button pressed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Sleep button released',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Reset button pressed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Reset button released',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'FRU latch open',
            'severity': const.Health.Ok,
            'deassertion_desc': 'FRU latch closed',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Service requested',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Service no longer requested',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x19: {  # chip set
        0: {
            'desc': 'Soft power control failure',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Soft power control failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Thermal trip',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Thermal trip deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x1b: {  # Cable/Interconnect
        0: {
            'desc': 'Connected',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Disconnected',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Connection error',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Connection error deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x1d: {  # system boot initiated
        0: {
            'desc': 'Power up',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Power up deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Hard reset',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Hard reset deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Warm reset',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Warm reset deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'PXE boot',
            'severity': const.Health.Ok,
            'deassertion_desc': 'PXE boot deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Automatic boot to diagnostic',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Automatic boot to diagnostic deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'OS hard reset',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS hard reset deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'OS warm reset',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS warm reset deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'System restart',
            'severity': const.Health.Ok,
            'deassertion_desc': 'System restart deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x1e: {  # boot error
        0: {
            'desc': 'No bootable media',
            'severity': const.Health.Failed,
            'deassertion_desc': 'No bootable media deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Unbootable removable media',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Unbootable removable media deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'PXE failure',
            'severity': const.Health.Failed,
            'deassertion_desc': 'PXE failure deasserted',
            'deassertion_severity': const.Health.Ok,

        },
        3: {
            'desc': 'Invalid boot sector',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Invalid boot sector deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Interactive boot timeout',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Interactive boot timeout deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x1f: {  # OS boot
        0: {
            'desc': 'A: boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Unbootable removable media deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Hard drive boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Hard drive boot completed not failed',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Network boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Network boot completion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Diagnostic boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Diagnostic boot deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'CD boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'CD boot completion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'ROM boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'ROM boot completion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Boot completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Boot completion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'OS deployment started',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS deployment deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        8: {
            'desc': 'OS deployment completed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS deployment completion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        9: {
            'desc': 'OS deployment aborted',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS deployment abortion deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xa: {
            'desc': 'OS deployment failed',
            'severity': const.Health.Failed,
            'deassertion_desc': 'OS deployment failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x20: {  # OS Stop
        0: {
            'desc': 'OS boot stop',
            'severity': const.Health.Failed,
            'deassertion_desc': 'OS boot stop deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'OS crash',
            'severity': const.Health.Failed,
            'deassertion_desc': 'OS crash deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'OS cleanly halted',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS clean halt deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'OS cleanly shutdown',
            'severity': const.Health.Ok,
            'deassertion_desc': 'OS clean shutdown deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Event driven soft shutdown',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Event driven soft shutdown deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Event driven soft shutdown failed',
            'severity': const.Health.Warning,
            'deassertion_desc':
            'Event driven soft shutdown failure deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x21:  {  # slot/connector
        0x0: {
            'desc': 'Fault',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Fault recovered',
            'deassertion_severity': const.Health.Ok,
        },
        0x1: {
            'desc': 'Identify',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Identify deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x2: {
            'desc': 'Slot/connector installed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Slot/connector removed',
            'deassertion_severity': const.Health.Ok,
        },
        0x3: {
            'desc': 'Slot/connector ready for install',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Slot/connector no longer ready for install',
            'deassertion_severity': const.Health.Ok,
        },
        0x4: {
            'desc': 'Slot/connector ready for removal',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Slot/connector no longer ready for removal',
            'deassertion_severity': const.Health.Ok,
        },
        0x5: {
            'desc': 'Slot powered down',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Slot powered up',
            'deassertion_severity': const.Health.Ok,
        },
        0x6: {
            'desc': 'Slot/connector device removal requested',
            'severity': const.Health.Warning,
            'deassertion_desc':
            'Slot/connector device removal no longer requested',
            'deassertion_severity': const.Health.Ok,
        },
        0x7: {
            'desc': 'Slot/connector interlock',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Slot/connector interlock deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x8: {
            'desc': 'Slot/connector disabled',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Slot/connector enabled',
            'deassertion_severity': const.Health.Ok,
        },
        0x9: {
            'desc': 'Slot holds spare device',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Slot no longer holds spare device',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x22: {  # system acpi power state
        0x0: {
            'desc': 'Online',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Online deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x1: {
            'desc': 'S1 sleep',
            'severity': const.Health.Ok,
            'deassertion_desc': 'S1 sleep deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x2: {
            'desc': 'S2 sleep',
            'severity': const.Health.Ok,
            'deassertion_desc': 'S2 sleep deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x3: {
            'desc': 'Sleep',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Sleep deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x4: {
            'desc': 'Hibernated',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Hibernation deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x5: {
            'desc': 'Off',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Off deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x6: {
            'desc': 'Hibernated or off',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Hibernate or off deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x7: {
            'desc': 'Mechanically off',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Mechanically on',
            'deassertion_severity': const.Health.Ok,
        },
        0x8: {
            'desc': 'Sleep',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Sleep deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x9: {
            'desc': 'G1 sleep',
            'severity': const.Health.Ok,
            'deassertion_desc': 'G1 sleep deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xa: {
            'desc': 'Shutdown',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Shutdown deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0xb: {
            'desc': 'On',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Off',
            'deassertion_severity': const.Health.Ok,
        },
        0xc: {
            'desc': 'Off',
            'severity': const.Health.Ok,
            'deassertion_desc': 'On',
            'deassertion_severity': const.Health.Ok,
        },
        0xe: {
            'desc': 'Unknown',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Unknown deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x23: {  # watchdog
        0x0: {
            'desc': 'Watchdog expired',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Watchdog expiry deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x1: {
            'desc': 'Watchdog hard reset',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Watchdog hard reset deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x2: {
            'desc': 'Watchdog power down',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Watchdog power down deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x3: {
            'desc': 'Watchdog power cycle',
            'severity': const.Health.Failed,
            'deassertion_desc': 'Watchdog power cycle deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x8: {
            'desc': 'Watchdog interrupt',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Watchdog interrupt deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x24: {  # platform event
        0x0: {
            'desc': 'Platform generated page',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Platform generated page deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x1: {
            'desc': 'Platform generated network alert',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Platform generated network alert deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x2: {
            'desc': 'Platform generated PET alert',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Platform generated PET alert deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        0x3: {
            'desc': 'Platform generated OEM SNMP Alert',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Platform generated OEM SNMP Alert deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x25: {  # entity presence
        0: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Absent',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Present',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Disabled',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Enabled',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x27: {  # LAN heartbeat
        0: {
            'desc': 'Heartbeat lost',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Heartbeat recovered',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Heartbeat',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Heartbeat lost',
            'deassertion_severity': const.Health.Warning,
        },
    },
    0x28: {  # management subsystem health
        0: {
            'desc': 'Sensor access degraded or unavailable',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Sensor access restored',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Controller access degraded or unavailable',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Controller access restored',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Controller Offline',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Controller online',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Controller Offline',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Controller online',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Sensor Error',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Sensor recovered',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'FRU Failure',
            'severity': const.Health.Warning,
            'deassertion_desc': 'FRU recovered',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x29: {  # battery
        0: {
            'desc': 'Low',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Battery no longer low',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Failed',
            # Critical here because typical battery failure
            # does not indicate a 'failed' runtime
            'severity': const.Health.Critical,
            'deassertion_desc': 'Recovered',
            'deassertion_severity': const.Health.Ok
        },
        2: {
            'desc': 'Present',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Absent',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x2a: {  # session audit
        0: {
            'desc': 'Activated',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Deactivated',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Deactivated',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Activated',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Invalid username/password',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Valid username/password',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Account disabled due to failure count',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Account re-enabled',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x2b: {  # Version Change
        0: {
            'desc': 'Hardware change detected',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Hardware change deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Firmware or software change detected',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Firmware or software change deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Hardware incompatibility detected',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Hardware incompatibility deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Firmware/software incompatibility detected',
            'severity': const.Health.Critical,
            'deassertion_desc': 'Firmware/software incompatibility deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Invalid/Unsupported hardware revision',
            'severity': const.Health.Critical,
            'deassertion_desc':
            'Invalid/unsupported hardware revision deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Invalid/unsupported firmware/software version',
            'severity': const.Health.Critical,
            'deassertion_desc':
            'Invalid/unsupported firmware/software version',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Successful hardware change',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Successful hardware change deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Successful software/firmware change',
            'severity': const.Health.Ok,
            'deassertion_desc':
            'Successful software/firmware change deasserted',
            'deassertion_severity': const.Health.Ok,
        },
    },
    0x2c: {  # FRU state
        0: {
            'desc': 'Not Installed',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Installed',
            'deassertion_severity': const.Health.Ok,
        },
        1: {
            'desc': 'Inactive',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Inactive deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        2: {
            'desc': 'Activation requested',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Activation request deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        3: {
            'desc': 'Activation in progress',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Activation no longer in progress',
            'deassertion_severity': const.Health.Ok,
        },
        4: {
            'desc': 'Active',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Inactive',
            'deassertion_severity': const.Health.Ok,
        },
        5: {
            'desc': 'Deactivation requested',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Deactivation request deasserted',
            'deassertion_severity': const.Health.Ok,
        },
        6: {
            'desc': 'Deactivation in progress',
            'severity': const.Health.Ok,
            'deassertion_desc': 'Deactivation no longer in progress',
            'deassertion_severity': const.Health.Ok,
        },
        7: {
            'desc': 'Communication lost',
            'severity': const.Health.Warning,
            'deassertion_desc': 'Communication restored',
            'deassertion_severity': const.Health.Ok,
        },
    },
}

# entity ids from table 43-13 entity id codes
entity_ids = {
    0x0: 'Unspecified',
    0x1: 'Other',
    0x2: 'Unknown',
    0x3: 'Processor',
    0x4: 'Disk or disk bay',
    0x5: 'Peripheral bay',
    0x6: 'System management module',
    0x7: 'System board',
    0x8: 'Memory module',
    0x9: 'Processor module',
    0xa: 'Power supply',
    0xb: 'Add-in card',
    0xc: 'Front panel board',
    0xd: 'Back panel board',
    0xe: 'Power system board',
    0xf: 'Drive backplane',
    0x10: 'System internal expansion board',
    0x11: 'Other system board',
    0x12: 'Processor board',
    0x13: 'Power unit / power domain',
    0x14: 'Power module / DC-to-DC converter',
    0x15: 'Power management /power distribution board',
    0x16: 'Chassis back panel board',
    0x17: 'System chassis',
    0x18: 'Sub-chassis',
    0x19: 'Other chassis board',
    0x1a: 'Disk drive bay',
    0x1b: 'Peripheral bay',
    0x1c: 'Device bay',
    0x1d: 'Fan/cooling device',
    0x1e: 'Cooling unit / cooling domain',
    0x1f: 'Cable / interconnect',
    0x20: 'Memory device',
    0x21: 'System management software',
    0x22: 'System firmware',
    0x23: 'Operating system',
    0x24: 'System bus',
    0x25: 'Group',
    0x26: 'Remote management communication device',
    0x27: 'External environment',
    0x28: 'Battery',
    0x29: 'Processing blade',
    0x2a: 'Connectivity switch',
    0x2b: 'Processor/memory module',
    0x2c: 'I/O module',
    0x2d: 'Processor I/O module',
    0x2e: 'Management controller firmware',
    0x2f: 'IPMI channel',
    0x30: 'PCI Bus',
    0x31: 'PCIe Bus',
    0x32: 'SCSI Bus',
    0x33: 'SATA/SAS Bus',
    0x34: 'Processor / front-side bus',
    0x35: 'Real time clock',
    0x37: 'Air inlet',
    0x40: 'Air inlet',
    0x41: 'Processor',
    0x42: 'System board',
}


rmcp_codes = {
    1: ("Insufficient resources to create new session (wait for existing "
        "sessions to timeout)"),
    2: "Invalid Session ID",
    3: "Invalid payload type",
    4: "Invalid authentication algorithm",
    5: "Invalid integrity algorithm",
    6: "No matching integrity payload",
    7: "No matching integrity payload",
    8: "Inactive Session ID",
    9: "Invalid role",
    0xa: "Unauthorized role or privilege level requested",
    0xb: "Insufficient resources to create a session at the requested role",
    0xc: "Invalid username length",
    0xd: "Unauthorized name",
    0xe: "Unauthorized GUID",
    0xf: "Invalid integrity check value",
    0x10: "Invalid confidentiality algorithm",
    0x11: "No Cipher suite match with proposed security algorithms",
    0x12: "Illegal or unrecognized parameter",
}

netfn_codes = {
    "chassis": 0x0,
    "bridge": 0x2,
    "sensorevent": 0x4,
    "application": 0x6,
    "firmware": 0x8,
    "storage": 0xa,
    "transport": 0xc,
}

command_completion_codes = {
    (7, 0x39): {
        0x81: "Invalid user name",
        0x82: "Null user disabled",
    },
    (7, 0x3a): {
        0x81: "No available login slots",
        0x82: "No available login slots for requested user",
        0x83: "No slot available with requested privilege level",
        0x84: "Session sequence number out of range",
        0x85: "Invalid session ID",
        0x86: ("Requested privilege level exceeds requested user permissions "
               "on this channel"),
    },
    (7, 0x3b): {  # Set session privilege level
        0x80: "User is not allowed requested privilege level",
        0x81: "Requested privilege level is not allowed over this channel",
        0x82: "Cannot disable user level authentication",
    },
    (1, 8): {  # set system boot options
        0x80: "Parameter not supported",
        0x81: "Attempt to set set 'set in progress' when not 'set complete'",
        0x82: "Attempt to write read-only parameter",
    },
    (7, 0x48): {  # activate payload
        0x80: "Payload already active on another session",
        0x81: "Payload is disabled",
        0x82: "Payload activation limit reached",
        0x83: "Cannot activate payload with encryption",
        0x84: "Cannot activate payload without encryption",
    },
    (6, 0x47): {  # set user password
        0x80: "Password test failed. Password does not match stored value",
        0x81: "Password test failed. Wrong password size was used"
    },
}

ipmi_completion_codes = {
    0x00: "Success",
    0xc0: "Node Busy",
    0xc1: "Invalid command",
    0xc2: "Invalid command for given LUN",
    0xc3: "Timeout while processing command",
    0xc4: "Out of storage space on BMC",
    0xc5: "Reservation canceled or invalid reservation ID",
    0xc6: "Request data truncated",
    0xc7: "Request data length invalid",
    0xc8: "Request data field length limit exceeded",
    0xc9: "Parameter out of range",
    0xca: "Cannot return number of requested data bytes",
    0xcb: "Requested sensor, data, or record not present",
    0xcc: "Invalid data field in request",
    0xcd: "Command illegal for specified sensor or record type",
    0xce: "Command response could not be provided",
    0xcf: "Cannot execute duplicated request",
    0xd0: "SDR repository in update mode",
    0xd1: "Device in firmware update mode",
    0xd2: "BMC initialization in progress",
    0xd3: "Internal destination unavailable",
    0xd4: "Insufficient privilege level or firmware firewall",
    0xd5: "Command not supported in present state",
    0xd6: "Cannot execute command because subfunction disabled or unavailable",
    0xff: "Unspecified",
    0xffff: "Timeout",  # not ipmi, but used internally
}



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2016 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from datetime import datetime
import json
import pyghmi.ipmi.private.util as util
import pyghmi.util.webclient as webclient
import urllib
import weakref


class IMMClient(object):
    logouturl = '/data/logout'
    bmcname = 'IMM'

    def __init__(self, ipmicmd):
        self.ipmicmd = weakref.proxy(ipmicmd)
        self.imm = ipmicmd.bmc
        self.username = ipmicmd.ipmi_session.userid
        self.password = ipmicmd.ipmi_session.password
        self._wc = None  # The webclient shall be initiated on demand
        self.datacache = {}

    @staticmethod
    def _parse_builddate(strval):
        try:
            return datetime.strptime(strval, '%Y/%m/%d %H:%M:%S')
        except ValueError:
            pass
        try:
            return datetime.strptime(strval, '%Y-%m-%d %H:%M:%S')
        except ValueError:
            pass
        try:
            return datetime.strptime(strval, '%Y/%m/%d')
        except ValueError:
            pass
        try:
            return datetime.strptime(strval, '%m/%d/%Y')
        except ValueError:
            pass
        try:
            return datetime.strptime(strval, '%Y-%m-%d')
        except ValueError:
            pass
        try:
            return datetime.strptime(strval, '%m %d %Y')
        except ValueError:
            pass
        return None

    @classmethod
    def parse_imm_buildinfo(cls, buildinfo):
        buildid = buildinfo[:9].rstrip(' \x00')
        bdt = ' '.join(buildinfo[9:].replace('\x00', ' ').split())
        bdate = cls._parse_builddate(bdt)
        return buildid, bdate

    @classmethod
    def datefromprop(cls, propstr):
        if propstr is None:
            return None
        return cls._parse_builddate(propstr)

    def get_property(self, propname):
        propname = propname.encode('utf-8')
        proplen = len(propname) | 0b10000000
        cmdlen = len(propname) + 1
        cdata = bytearray([0, 0, cmdlen, proplen]) + propname
        rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0xc4, data=cdata)
        rsp['data'] = bytearray(rsp['data'])
        if rsp['data'][0] != 0:
            return None
        propdata = rsp['data'][3:]  # second two bytes are size, don't need it
        if propdata[0] & 0b10000000:  # string, for now assume length valid
            return str(propdata[1:]).rstrip(' \x00')
        else:
            raise Exception('Unknown format for property: ' + repr(propdata))

    def get_webclient(self):
        cv = self.ipmicmd.certverify
        wc = webclient.SecureHTTPConnection(self.imm, 443, verifycallback=cv)
        try:
            wc.connect()
        except Exception:
            return None
        adata = urllib.urlencode({'user': self.username,
                                  'password': self.password,
                                  'SessionTimeout': 60
                                  })
        headers = {'Connection': 'keep-alive',
                   'Referer': 'https://{0}/designs/imm/index.php'.format(
                       self.imm),
                   'Content-Type': 'application/x-www-form-urlencoded'}
        wc.request('POST', '/data/login', adata, headers)
        rsp = wc.getresponse()
        if rsp.status == 200:
            rspdata = json.loads(rsp.read())
            if rspdata['authResult'] == '0' and rspdata['status'] == 'ok':
                if 'token2_name' in rspdata and 'token2_value' in rspdata:
                    wc.set_header(rspdata['token2_name'],
                                  rspdata['token2_value'])
                return wc

    @property
    def wc(self):
        if not self._wc:
            self._wc = self.get_webclient()
        return self._wc

    def fetch_grouped_properties(self, groupinfo):
        retdata = {}
        for keyval in groupinfo:
            retdata[keyval] = self.get_property(groupinfo[keyval])
            if keyval == 'date':
                retdata[keyval] = self.datefromprop(retdata[keyval])
        returnit = False
        for keyval in list(retdata):
            if retdata[keyval] in (None, ''):
                del retdata[keyval]
            else:
                returnit = True
        if returnit:
            return retdata

    def get_cached_data(self, attribute):
        try:
            kv = self.datacache[attribute]
            if kv[1] > util._monotonic_time() - 30:
                return kv[0]
        except KeyError:
            return None

    def attach_remote_media(self, url, user, password):
        url = url.replace(':', '\:')
        params = urllib.urlencode({
            'RP_VmAllocateMountUrl({0},{1},1,,)'.format(
                self.username, url): ''
        })
        result = self.wc.grab_json_response('/data?set', params)
        if result['return'] != 'Success':
            raise Exception(result['reason'])
        self.weblogout()

    def detach_remote_media(self):
        mnt = self.wc.grab_json_response(
            '/designs/imm/dataproviders/imm_rp_images.php')
        removeurls = []
        for item in mnt['items']:
            if 'urls' in item:
                for url in item['urls']:
                    removeurls.append(url['url'])
        for url in removeurls:
            url = url.replace(':', '\:')
            params = urllib.urlencode({
                'RP_VmAllocateUnMountUrl({0},{1},0,)'.format(
                    self.username, url): ''})
            result = self.wc.grab_json_response('/data?set', params)
            if result['return'] != 'Success':
                raise Exception(result['reason'])
        self.weblogout()

    def fetch_agentless_firmware(self):
        adapterdata = self.get_cached_data('lenovo_cached_adapters')
        if not adapterdata:
            if self.wc:
                adapterdata = self.wc.grab_json_response(
                    '/designs/imm/dataproviders/imm_adapters.php')
                if adapterdata:
                    self.datacache['lenovo_cached_adapters'] = (
                        adapterdata, util._monotonic_time())
        if adapterdata and 'items' in adapterdata:
            for adata in adapterdata['items']:
                aname = adata['adapter.adapterName']
                donenames = set([])
                for fundata in adata['adapter.functions']:
                    fdata = fundata.get('firmwares', ())
                    for firm in fdata:
                        fname = firm['firmwareName'].rstrip()
                        if '.' in fname:
                            fname = firm['description'].rstrip()
                        if fname in donenames:
                            # ignore redundant entry
                            continue
                        donenames.add(fname)
                        bdata = {}
                        if 'versionStr' in firm and firm['versionStr']:
                            bdata['version'] = firm['versionStr']
                        if ('releaseDate' in firm and
                                firm['releaseDate'] and
                                firm['releaseDate'] != 'N/A'):
                            try:
                                bdata['date'] = self._parse_builddate(
                                    firm['releaseDate'])
                            except ValueError:
                                pass
                        yield ('{0} {1}'.format(aname, fname), bdata)
        storagedata = self.get_cached_data('lenovo_cached_storage')
        if not storagedata:
            if self.wc:
                storagedata = self.wc.grab_json_response(
                    '/designs/imm/dataproviders/raid_alldevices.php')
                if storagedata:
                    self.datacache['lenovo_cached_storage'] = (
                        storagedata, util._monotonic_time())
        if storagedata and 'items' in storagedata:
            for adp in storagedata['items']:
                if 'storage.vpd.productName' not in adp:
                    continue
                adpname = adp['storage.vpd.productName']
                if 'children' not in adp:
                    adp['children'] = ()
                for diskent in adp['children']:
                    bdata = {}
                    diskname = '{0} Disk {1}'.format(
                        adpname,
                        diskent['storage.slotNo'])
                    bdata['model'] = diskent[
                        'storage.vpd.productName'].rstrip()
                    bdata['version'] = diskent['storage.firmwares'][0][
                        'versionStr']
                    yield (diskname, bdata)
        self.weblogout()

    def get_hw_inventory(self):
        hwmap = self.hardware_inventory_map()
        for key in hwmap:
            yield (key, hwmap[key])

    def get_hw_descriptions(self):
        hwmap = self.hardware_inventory_map()
        for key in hwmap:
            yield key

    def get_component_inventory(self, compname):
        hwmap = self.hardware_inventory_map()
        try:
            return hwmap[compname]
        except KeyError:
            return None

    def weblogout(self):
        if self._wc:
            self._wc.grab_json_response(self.logouturl)
            self._wc = None

    def hardware_inventory_map(self):
        hwmap = self.get_cached_data('lenovo_cached_hwmap')
        if hwmap:
            return hwmap
        hwmap = {}
        adapterdata = self.get_cached_data('lenovo_cached_adapters')
        if not adapterdata:
            if self.wc:
                adapterdata = self.wc.grab_json_response(
                    '/designs/imm/dataproviders/imm_adapters.php')
                if adapterdata:
                    self.datacache['lenovo_cached_adapters'] = (
                        adapterdata, util._monotonic_time())
        if adapterdata and 'items' in adapterdata:
            for adata in adapterdata['items']:
                skipadapter = False
                if not adata['adapter.oobSupported']:
                    continue
                aname = adata['adapter.adapterName']
                clabel = adata['adapter.connectorLabel']
                if clabel == 'Unknown':
                    continue
                if clabel != 'Onboard':
                    aslot = adata['adapter.slotNo']
                    if clabel == 'ML2':
                        clabel = 'ML2 (Slot {0})'.format(aslot)
                    else:
                        clabel = 'Slot {0}'.format(aslot)
                bdata = {'location': clabel}
                for fundata in adata['adapter.functions']:
                    bdata['pcislot'] = '{0:02x}:{1:02x}'.format(
                        fundata['generic.busNo'], fundata['generic.devNo'])
                    serialdata = fundata.get('vpd.serialNo', None)
                    if (serialdata and serialdata != 'N/A' and
                            '---' not in serialdata):
                        bdata['serial'] = serialdata
                    partnum = fundata.get('vpd.partNo', None)
                    if partnum and partnum != 'N/A':
                        bdata['partnumber'] = partnum
                    if 'network.pPorts' in fundata:
                        for portinfo in fundata['network.pPorts']:
                            for lp in portinfo['logicalPorts']:
                                ma = lp['networkAddr']
                                ma = ':'.join(
                                    [ma[i:i+2] for i in range(
                                        0, len(ma), 2)]).lower()
                                bdata['MAC Address {0}'.format(
                                    portinfo['portIndex'])] = ma
                    elif clabel == 'Onboard':  # skip the various non-nic
                        skipadapter = True
                if not skipadapter:
                    hwmap[aname] = bdata
            self.datacache['lenovo_cached_hwmap'] = (hwmap,
                                                     util._monotonic_time())
        self.weblogout()
        return hwmap

    def get_firmware_inventory(self, bmcver):
        # First we fetch the system firmware found in imm properties
        # then check for agentless, if agentless, get adapter info using
        # https, using the caller TLS verification scheme
        rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0x50)
        immverdata = self.parse_imm_buildinfo(rsp['data'])
        bdata = {
            'version': bmcver, 'build': immverdata[0], 'date': immverdata[1]}
        yield (self.bmcname, bdata)
        bdata = self.fetch_grouped_properties({
            'build': '/v2/ibmc/dm/fw/imm2/backup_build_id',
            'version': '/v2/ibmc/dm/fw/imm2/backup_build_version',
            'date': '/v2/ibmc/dm/fw/imm2/backup_build_date'})
        if bdata:
            yield ('{0} Backup'.format(self.bmcname), bdata)
            bdata = self.fetch_grouped_properties({
                'build': '/v2/ibmc/trusted_buildid',
            })
        if bdata:
            yield ('{0} Trusted Image'.format(self.bmcname), bdata)
        bdata = self.fetch_grouped_properties({
            'build': '/v2/bios/build_id',
            'version': '/v2/bios/build_version',
            'date': '/v2/bios/build_date'})
        if bdata:
            yield ('UEFI', bdata)
        else:
            yield ('UEFI', {'version': 'unknown'})
        bdata = self.fetch_grouped_properties({
            'build': '/v2/ibmc/dm/fw/bios/backup_build_id',
            'version': '/v2/ibmc/dm/fw/bios/backup_build_version'})
        if bdata:
            yield ('UEFI Backup', bdata)
        # Note that the next pending could be pending for either primary
        # or backup, so can't promise where it will go
        bdata = self.fetch_grouped_properties({
            'build': '/v2/bios/pending_build_id'})
        if bdata:
            yield ('UEFI Pending Update', bdata)
        fpga = self.ipmicmd.xraw_command(netfn=0x3a, command=0x6b, data=(0,))
        fpga = '{0}.{1}.{2}'.format(*[ord(x) for x in fpga['data']])
        yield ('FPGA', {'version': fpga})
        for firm in self.fetch_agentless_firmware():
            yield firm


class XCCClient(IMMClient):
    logouturl = '/api/providers/logout'
    bmcname = 'XCC'

    def get_webclient(self):
        cv = self.ipmicmd.certverify
        wc = webclient.SecureHTTPConnection(self.imm, 443, verifycallback=cv)
        try:
            wc.connect()
        except Exception:
            return None
        adata = json.dumps({'username': self.username,
                            'password': self.password
                            })
        headers = {'Connection': 'keep-alive',
                   'Content-Type': 'application/json'}
        wc.request('POST', '/api/login', adata, headers)
        rsp = wc.getresponse()
        if rsp.status == 200:
            rspdata = json.loads(rsp.read())
            wc.set_header('Content-Type', 'application/json')
            wc.set_header('Authorization', 'Bearer ' + rspdata['access_token'])
            if '_csrf_token' in wc.cookies:
                wc.set_header('X-XSRF-TOKEN', wc.cookies['_csrf_token'])
            return wc

    def attach_remote_media(self, url, user, password):
        proto, host, path = util.urlsplit(url)
        if proto == 'smb':
            proto = 'cifs'
        rq = {'Option': '', 'Domain': '', 'Write': 0}
        # nfs == 1, cifs == 0
        if proto == 'nfs':
            rq['Protocol'] = 1
            rq['Url'] = '{0}:{1}'.format(host, path)
        elif proto == 'cifs':
            rq['Protocol'] = 0
            rq['Credential'] = '{0}:{1}'.format(user, password)
            rq['Url'] = '//{0}{1}'.format(host, path)
        elif proto in ('http', 'https'):
            rq['Protocol'] = 7
            rq['Url'] = url
        else:
            raise Exception('TODO')
        rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_connect',
                                        json.dumps(rq))
        if 'return' not in rt or rt['return'] != 0:
            raise Exception('Unhandled return: ' + repr(rt))
        rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_mountall',
                                        '{}')
        if 'return' not in rt or rt['return'] != 0:
            raise Exception('Unhandled return: ' + repr(rt))

    def get_firmware_inventory(self, bmcver):
        # First we fetch the system firmware found in imm properties
        # then check for agentless, if agentless, get adapter info using
        # https, using the caller TLS verification scheme
        rsp = self.ipmicmd.xraw_command(netfn=0x3a, command=0x50)
        immverdata = self.parse_imm_buildinfo(rsp['data'])
        bdata = {
            'version': bmcver, 'build': immverdata[0], 'date': immverdata[1]}
        yield (self.bmcname, bdata)
        bdata = self.fetch_grouped_properties({
            'build': '/v2/ibmc/dm/fw/imm3/backup_pending_build_id',
            'version': '/v2/ibmc/dm/fw/imm3/backup_pending_build_version',
            'date': '/v2/ibmc/dm/fw/imm3/backup_pending_build_date'})
        if bdata:
            yield ('{0} Backup'.format(self.bmcname), bdata)
        else:
            bdata = self.fetch_grouped_properties({
                'build': '/v2/ibmc/dm/fw/imm3/backup_build_id',
                'version': '/v2/ibmc/dm/fw/imm3/backup_build_version',
                'date': '/v2/ibmc/dm/fw/imm3/backup_build_date'})
            if bdata:
                yield ('{0} Backup'.format(self.bmcname), bdata)
                bdata = self.fetch_grouped_properties({
                    'build': '/v2/ibmc/trusted_buildid',
                })
        if bdata:
            bdata = self.fetch_grouped_properties({
                'build': '/v2/ibmc/trusted_buildid',
            })
        if bdata:
            yield ('{0} Trusted Image'.format(self.bmcname), bdata)
        bdata = self.fetch_grouped_properties({
            'build': '/v2/bios/build_id',
            'version': '/v2/bios/build_version',
            'date': '/v2/bios/build_date'})
        if bdata:
            yield ('UEFI', bdata)
        # Note that the next pending could be pending for either primary
        # or backup, so can't promise where it will go
        bdata = self.fetch_grouped_properties({
            'build': '/v2/bios/pending_build_id'})
        if bdata:
            yield ('UEFI Pending Update', bdata)
        bdata = self.fetch_grouped_properties({
            'build': '/v2/tdm/build_id',
            'version': '/v2/tdm/build_version',
            'date': '/v2/tdm/build_date'})
        if bdata:
            yield ('LXPM', bdata)
        fpga = self.ipmicmd.xraw_command(netfn=0x3a, command=0x6b, data=(0,))
        fpga = '{0}.{1}.{2}'.format(*[ord(x) for x in fpga['data']])
        yield ('FPGA', {'version': fpga})
        for firm in self.fetch_agentless_firmware():
            yield firm

    def detach_remote_media(self):
        rt = self.wc.grab_json_response('/api/providers/rp_vm_remote_getdisk')
        if 'items' in rt:
            slots = []
            for mount in rt['items']:
                slots.append(mount['slotId'])
            for slot in slots:
                rt = self.wc.grab_json_response(
                    '/api/providers/rp_vm_remote_unmount',
                    json.dumps({'Slot': slot}))
                if 'return' not in rt or rt['return'] != 0:
                    raise Exception("Unrecognized return: " + repr(rt))



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# coding=utf8

# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This module provides access to SDR offered by a BMC
# This data is common between 'sensors' and 'inventory' modules since SDR
# is both used to enumerate sensors for sensor commands and FRU ids for FRU
# commands

# For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
# code.  Will see if it is adequate to advocate for high object reuse in a
# persistent process for the moment.

# Focus is at least initially on the aspects that make the most sense for a
# remote client to care about.  For example, smbus information is being
# skipped for now

# This file handles parsing of fru format records as presented by IPMI
# devices.  This format is documented in the 'Platform Management FRU
# Information Storage Definition (Document Revision 1.2)

import pyghmi.exceptions as iexc
import pyghmi.ipmi.private.spd as spd
import struct
import time

fruepoch = 820454400  # 1/1/1996, 0:00

# This is from SMBIOS specification Table 16
enclosure_types = {
    1: 'Other',
    2: 'Unknown',
    3: 'Desktop',
    4: 'Low Profile Desktop',
    5: 'Pizza Box',
    6: 'Mini Tower',
    7: 'Tower',
    8: 'Portable',
    9: 'Laptop',
    0xa: 'Notebook',
    0xb: 'Hand Held',
    0xc: 'Docking Station',
    0xd: 'All in One',
    0xe: 'Sub Notebook',
    0xf: 'Space-saving',
    0x10: 'Lunch Box',
    0x11: 'Main Server Chassis',
    0x12: 'Expansion Chassis',
    0x13: 'SubChassis',
    0x14: 'Bus Expansion Chassis',
    0x15: 'Peripheral Chassis',
    0x16: 'RAID Chassis',
    0x17: 'Rack Mount Chassis',
    0x18: 'Sealed-case PC',
    0x19: 'Multi-system Chassis',
    0x1a: 'Compact PCI',
    0x1b: 'Advanced TCA',
    0x1c: 'Blade',
    0x1d: 'Blade Enclosure',
}


def unpack6bitascii(inputdata):
    # This is a text encoding scheme that seems unique
    # to IPMI FRU.  It seems to be relatively rare in practice
    result = ''
    while len(inputdata) > 0:
        currchunk = inputdata[:3]
        del inputdata[:3]
        currchar = currchunk[0] & 0b111111
        result += chr(0x20 + currchar)
        currchar = (currchunk[0] & 0b11000000) >> 6
        currchar |= (currchunk[1] & 0b1111) << 2
        result += chr(0x20 + currchar)
        currchar = (currchunk[1] & 0b11110000) >> 4
        currchar |= (currchunk[2] & 0b11) << 4
        result += chr(0x20 + currchar)
        currchar = (currchunk[2] & 0b11111100) >> 2
        result += chr(0x20 + currchar)
    return result


def decode_fru_date(datebytes):
    # Returns ISO
    datebytes.append(0)
    minutesfromepoch = struct.unpack('<I', struct.pack('4B', *datebytes))[0]
    # Some data in the field has had some data less than 800
    # At this juncture, it's far more likely for this noise
    # to be incorrect than anything in particular
    if minutesfromepoch < 800:
        return None
    return time.strftime('%Y-%m-%dT%H:%M',
                         time.gmtime((minutesfromepoch * 60) + fruepoch))


class FRU(object):
    """An object representing structure

    FRU (Field Replaceable Unit) is the usual format for inventory in IPMI
    devices.  This covers most standards compliant inventory data
    as well as presenting less well defined fields in a structured way.

    :param rawdata: A binary string/bytearray of raw data from BMC or dump
    :param ipmicmd: An ipmi command object to fetch data live
    :param fruid: The identifier number of the FRU
    :param sdr: The sdr locator entry to help clarify how to parse data
    """

    def __init__(self, rawdata=None, ipmicmd=None, fruid=0, sdr=None):
        self.rawfru = rawdata
        self.databytes = None
        self.info = None
        self.sdr = sdr
        if self.rawfru is not None:
            self.parsedata()
        elif ipmicmd is not None:
            self.ipmicmd = ipmicmd
            # Use the ipmicmd to fetch the data
            try:
                self.fetch_fru(fruid)
            except iexc.IpmiException as ie:
                if ie.ipmicode in (203, 129):
                    return
                raise
            self.parsedata()
        else:
            raise TypeError('Either rawdata or ipmicmd must be specified')

    def fetch_fru(self, fruid):
        response = self.ipmicmd.raw_command(
            netfn=0xa, command=0x10, data=[fruid])
        if 'error' in response:
            raise iexc.IpmiException(response['error'], code=response['code'])
        frusize = response['data'][0] | (response['data'][1] << 8)
        # In our case, we don't need to think too hard about whether
        # the FRU is word or byte, we just process what we get back in the
        # payload
        chunksize = 240
        # Selected as it is accomodated by most tested things
        # and many tested things broke after going much
        # bigger
        if chunksize > frusize:
            chunksize = frusize
        offset = 0
        self.rawfru = bytearray([])
        while chunksize:
            response = self.ipmicmd.raw_command(
                netfn=0xa, command=0x11, data=[fruid, offset & 0xff,
                                               offset >> 8, chunksize])
            if response['code'] in (201, 202):
                # if it was too big, back off and try smaller
                # Try just over half to mitigate the chance of
                # one request becoming three rather than just two
                if chunksize == 3:
                    raise iexc.IpmiException(response['error'])
                chunksize //= 2
                chunksize += 2
                continue
            elif 'error' in response:
                raise iexc.IpmiException(response['error'], response['code'])
            self.rawfru.extend(response['data'][1:])
            offset += response['data'][0]
            if response['data'][0] == 0:
                break
            if offset + chunksize > frusize:
                chunksize = frusize - offset

    def parsedata(self):
        self.info = {}
        rawdata = self.rawfru
        self.databytes = bytearray(rawdata)
        if self.sdr is not None:
            frutype = self.sdr.fru_type_and_modifier >> 8
            frusubtype = self.sdr.fru_type_and_modifier & 0xff
            if frutype > 0x10 or frutype < 0x8 or frusubtype not in (0, 1, 2):
                return
                # TODO(jjohnson2): strict mode to detect pyghmi and BMC
                # gaps
                # raise iexc.PyghmiException(
                #     'Unsupported FRU device: {0:x}h, {1:x}h'.format(frutype,
                #                                                    frusubtype
                #                                                    ))
            elif frusubtype == 1:
                self.myspd = spd.SPD(self.databytes)
                self.info = self.myspd.info
                return
        if self.databytes[0] != 1:
            return
            # TODO(jjohnson2): strict mode to flag potential BMC errors
            # raise iexc.BmcErrorException("Invalid/Unsupported FRU format")
        # Ignore the internal use even if present.
        self._parse_chassis()
        self._parse_board()
        self._parse_prod()
        # TODO(jjohnson2): Multi Record area

    def _decode_tlv(self, offset, lang=0):
        currtlv = self.databytes[offset]
        currlen = currtlv & 0b111111
        currtype = (currtlv & 0b11000000) >> 6
        retinfo = self.databytes[offset + 1:offset + currlen + 1]
        newoffset = offset + currlen + 1
        if currlen == 0:
            return None, newoffset
        if currtype == 0:
            # return it as a bytearray, not much to be done for it
            return retinfo, newoffset
        elif currtype == 3:  # text string
            # Sometimes BMCs have FRU data with 0xff termination
            # contrary to spec, but can be tolerated
            # also in case something null terminates, handle that too
            # strictly speaking, \xff should be a y with diaeresis, but
            # erring on the side of that not being very relevant in practice
            # to fru info, particularly the last values
            retinfo = retinfo.rstrip('\xff\x00 ')
            if lang in (0, 25):
                try:
                    retinfo = retinfo.decode('iso-8859-1')
                except (UnicodeError, LookupError):
                    pass
            else:
                try:
                    retinfo = retinfo.decode('utf-16le')
                except (UnicodeDecodeError, LookupError):
                    pass
            # Some things lie about being text.  Do the best we can by
            # removing trailing spaces and nulls like makes sense for text
            # and rely on vendors to workaround deviations in their OEM
            # module
            retinfo = retinfo.rstrip('\x00 ')
            return retinfo, newoffset
        elif currtype == 1:  # BCD 'plus'
            retdata = ''
            for byte in retinfo:
                byte = hex(byte).replace('0x', '').replace('a', ' ').replace(
                    'b', '-').replace('c', '.')
                retdata += byte
            retdata = retdata.strip()
            return retdata, newoffset
        elif currtype == 2:  # 6-bit ascii
            retinfo = unpack6bitascii(retinfo).strip()
            return retinfo, newoffset

    def _parse_chassis(self):
        offset = 8 * self.databytes[2]
        if offset == 0:
            return
        if self.databytes[offset] & 0b1111 != 1:
            raise iexc.BmcErrorException("Invallid/Unsupported chassis area")
        inf = self.info
        # ignore length field, just process the data
        inf['Chassis type'] = enclosure_types[self.databytes[offset + 2]]
        inf['Chassis part number'], offset = self._decode_tlv(offset + 3)
        inf['Chassis serial number'], offset = self._decode_tlv(offset)
        inf['chassis_extra'] = []
        self.extract_extra(inf['chassis_extra'], offset)

    def extract_extra(self, target, offset, language=0):
        try:
            while self.databytes[offset] != 0xc1:
                fielddata, offset = self._decode_tlv(offset, language)
                target.append(fielddata)
        except IndexError:
            # If we overrun the end due to malformed FRU,
            # return at least what decoded right
            return

    def _parse_board(self):
        offset = 8 * self.databytes[3]
        if offset == 0:
            return
        if self.databytes[offset] & 0b1111 != 1:
            raise iexc.BmcErrorException("Invalid/Unsupported board info area")
        inf = self.info
        language = self.databytes[offset + 2]
        inf['Board manufacture date'] = decode_fru_date(
            self.databytes[offset + 3:offset + 6])
        inf['Board manufacturer'], offset = self._decode_tlv(offset + 6)
        inf['Board product name'], offset = self._decode_tlv(offset, language)
        inf['Board serial number'], offset = self._decode_tlv(offset, language)
        inf['Board model'], offset = self._decode_tlv(offset, language)
        _, offset = self._decode_tlv(offset, language)  # decode but discard
        inf['board_extra'] = []
        self.extract_extra(inf['board_extra'], offset, language)

    def _parse_prod(self):
        offset = 8 * self.databytes[4]
        if offset == 0:
            return
        inf = self.info
        language = self.databytes[offset + 2]
        inf['Manufacturer'], offset = self._decode_tlv(offset + 3,
                                                       language)
        inf['Product name'], offset = self._decode_tlv(offset, language)
        inf['Model'], offset = self._decode_tlv(offset, language)
        inf['Hardware Version'], offset = self._decode_tlv(offset, language)
        inf['Serial Number'], offset = self._decode_tlv(offset, language)
        inf['Asset Number'], offset = self._decode_tlv(offset, language)
        _, offset = self._decode_tlv(offset, language)
        inf['product_extra'] = []
        self.extract_extra(inf['product_extra'], offset, language)

    def __repr__(self):
        return repr(self.info)
        # retdata = 'Chassis data\n'
        # retdata += '   Type: ' + repr(self.chassis_type) + '\n'
        # retdata += '   Part Number: ' + repr(self.chassis_part_number) + '\n'
        # retdata += '   Serial Number: ' + repr(self.chassis_serial) + '\n'
        # retdata += '   Extra: ' + repr(self.chassis_extra) + '\n'
        # retdata += 'Board data\n'
        # retdata += '  Manufacturer: ' + repr(self.board_manufacturer) + '\n'
        # retdata += '   Date: ' + repr(self.board_mfg_date) + '\n'
        # retdata += '   Product' + repr(self.board_product) + '\n'
        # retdata += '   Serial: ' + repr(self.board_serial) + '\n'
        # retdata += '   Model: ' + repr(self.board_model) + '\n'
        # retdata += '   Extra: ' + repr(self.board_extra) + '\n'
        # retdata += 'Product data\n'
        # retdata += '  Manufacturer: ' + repr(self.product_manufacturer)+'\n'
        # retdata += '  Name: ' + repr(self.product_name) + '\n'
        # retdata += '  Model: ' + repr(self.product_model) + '\n'
        # retdata += '  Version: ' + repr(self.product_version) + '\n'
        # retdata += '  Serial: ' + repr(self.product_serial) + '\n'
        # retdata += '  Asset: ' + repr(self.product_asset) + '\n'
        # retdata += '  Extra: ' + repr(self.product_extra) + '\n'
        # return retdata



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2016-2017 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pyghmi.constants as pygconst
import pyghmi.exceptions as pygexc
import pyghmi.ipmi.sdr as sdr
import struct

try:
    range = xrange
except NameError:
    pass


def fpc_read_ac_input(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x90, data=(1,))
    rsp = rsp['data']
    if len(rsp) == 6:
        rsp = b'\x00' + bytes(rsp)
    return struct.unpack_from('<H', rsp[3:5])[0]


def fpc_read_dc_output(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x90, data=(2,))
    rsp = rsp['data']
    if len(rsp) == 6:
        rsp = b'\x00' + bytes(rsp)
    return struct.unpack_from('<H', rsp[3:5])[0]


def fpc_read_fan_power(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x90, data=(3,))
    rsp = rsp['data']
    rsp += '\x00'
    return struct.unpack_from('<I', rsp[1:])[0] / 100.0


def fpc_read_psu_fan(ipmicmd, number, sz):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa5, data=(number,))
    rsp = rsp['data']
    return struct.unpack_from('<H', rsp[:2])[0]


def fpc_get_psustatus(ipmicmd, number, sz):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0x91)
    mask = 1 << (number - 1)
    if len(rsp['data']) == 6:
        statdata = bytearray([0])
    else:
        statdata = bytearray()
    statdata += bytearray(rsp['data'])
    presence = statdata[3] & mask == mask
    pwrgood = statdata[4] & mask == mask
    throttle = (statdata[6] | statdata[2]) & mask == mask
    health = pygconst.Health.Ok
    states = []
    if presence and not pwrgood:
        health = pygconst.Health.Critical
        states.append('Power input lost')
    if throttle:
        health = pygconst.Health.Critical
        states.append('Throttled')
    if presence:
        states.append('Present')
    else:
        states.append('Absent')
        health = pygconst.Health.Critical
    return (health, states)


def fpc_get_nodeperm(ipmicmd, number, sz):
    try:
        rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa7, data=(number,))
    except pygexc.IpmiException as ie:
        if ie.ipmicode == 0xd5:  # no node present
            return (pygconst.Health.Ok, ['Absent'])
        raise
    perminfo = ord(rsp['data'][1])
    health = pygconst.Health.Ok
    states = []
    if len(rsp['data']) == 4:  # different gens handled rc differently
        rsp['data'] = b'\x00' + bytes(rsp['data'])
    if sz == 6:  # FPC
        permfail = ('\x02', '\x03')
    elif sz == 2:  # SMM
        permfail = ('\x02',)
    if rsp['data'][4] in permfail:
        states.append('Insufficient Power')
        health = pygconst.Health.Failed
    if perminfo & 0x40:
        states.append('Node Fault')
        health = pygconst.Health.Failed
    return (health, states)


def fpc_read_powerbank(ipmicmd):
    rsp = ipmicmd.xraw_command(netfn=0x32, command=0xa2)
    return struct.unpack_from('<H', rsp['data'][3:])[0]


fpc_sensors = {
    'AC Power': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_ac_input,
    },
    'DC Power': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_dc_output,
    },
    'Fan Power': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_fan_power
    },
    'PSU Fan Speed': {
        'type': 'Fan',
        'units': 'RPM',
        'provider': fpc_read_psu_fan,
        'elements': 1,
    },
    'Total Power Capacity': {
        'type': 'Power',
        'units': 'W',
        'provider': fpc_read_powerbank,
    },
    'Node Power Permission': {
        'type': 'Management Subsystem Health',
        'returns': 'tuple',
        'units': None,
        'provider': fpc_get_nodeperm,
        'elements': 2,
    },
    'Power Supply': {
        'type': 'Power Supply',
        'returns': 'tuple',
        'units': None,
        'provider': fpc_get_psustatus,
        'elements': 1,
    }
}


def get_sensor_names(size):
    global fpc_sensors
    for name in fpc_sensors:
        if size == 2 and name in ('Fan Power', 'Total Power Capacity'):
            continue
        sensor = fpc_sensors[name]
        if 'elements' in sensor:
            for elemidx in range(sensor['elements'] * size):
                elemidx += 1
                yield '{0} {1}'.format(name, elemidx)
        else:
            yield name


def get_sensor_descriptions(size):
    global fpc_sensors
    for name in fpc_sensors:
        if size == 2 and name in ('Fan Power', 'Total Power Capacity'):
            continue
        sensor = fpc_sensors[name]
        if 'elements' in sensor:
            for elemidx in range(sensor['elements'] * size):
                elemidx += 1
                yield {'name': '{0} {1}'.format(name, elemidx),
                       'type': sensor['type']}
        else:
            yield {'name': name, 'type': sensor['type']}


def get_fpc_firmware(bmcver, ipmicmd, fpcorsmm):
    mymsg = ipmicmd.xraw_command(netfn=0x32, command=0xa8)
    builddata = bytearray(mymsg['data'])
    name = None
    if fpcorsmm == 2:  # SMM
        name = 'SMM'
        buildid = '{0}{1}{2}{3}{4}{5}{6}'.format(
            *[chr(x) for x in builddata[-7:]])
    elif len(builddata) == 8:
        builddata = builddata[1:]  # discard the 'completion code'
        name = 'FPC'
        buildid = '{0}{1}'.format(builddata[-2], chr(builddata[-1]))
    yield (name, {'version': bmcver, 'build': buildid})
    yield ('PSOC', {'version': '{0}.{1}'.format(builddata[2], builddata[3])})


def get_sensor_reading(name, ipmicmd, sz):
    value = None
    sensor = None
    health = pygconst.Health.Ok
    states = []
    if name in fpc_sensors and 'elements' not in fpc_sensors[name]:
        sensor = fpc_sensors[name]
        value = sensor['provider'](ipmicmd)
    else:
        bnam, _, idx = name.rpartition(' ')
        idx = int(idx)
        if bnam in fpc_sensors and idx <= fpc_sensors[bnam]['elements'] * sz:
            sensor = fpc_sensors[bnam]
            if 'returns' in sensor:
                health, states = sensor['provider'](ipmicmd, idx, sz)
            else:
                value = sensor['provider'](ipmicmd, idx, sz)
    if sensor is not None:
        return sdr.SensorReading({'name': name, 'imprecision': None,
                                  'value': value, 'states': states,
                                  'state_ids': [], 'health': health,
                                  'type': sensor['type']},
                                 sensor['units'])
    raise Exception('Sensor not found: ' + name)



# vim: tabstop=4 shiftwidth=4 softtabstop=4
# coding=utf8

# Copyright 2014 IBM Corporation
# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This module provides access to SDR offered by a BMC
# This data is common between 'sensors' and 'inventory' modules since SDR
# is both used to enumerate sensors for sensor commands and FRU ids for FRU
# commands

# For now, we will not offer persistent SDR caching as we do in xCAT's IPMI
# code.  Will see if it is adequate to advocate for high object reuse in a
# persistent process for the moment.

# Focus is at least initially on the aspects that make the most sense for a
# remote client to care about.  For example, smbus information is being
# skipped for now

import math
import pyghmi.constants as const
import pyghmi.exceptions as exc
import pyghmi.ipmi.private.constants as ipmiconst
import struct
import weakref

TYPE_UNKNOWN = 0
TYPE_SENSOR = 1
TYPE_FRU = 2


def ones_complement(value, bits):
    # utility function to help with the large amount of 2s
    # complement prevalent in ipmi spec
    signbit = 0b1 << (bits - 1)
    if value & signbit:
        # if negative, subtract 1, then take 1s
        # complement given bits width
        return 0 - (value ^ ((0b1 << bits) - 1))
    else:
        return value


def twos_complement(value, bits):
    # utility function to help with the large amount of 2s
    # complement prevalent in ipmi spec
    signbit = 0b1 << (bits - 1)
    if value & signbit:
        # if negative, subtract 1, then take 1s
        # complement given bits width
        return 0 - ((value - 1) ^ ((0b1 << bits) - 1))
    else:
        return value


unit_types = {
    # table 43-15 'sensor unit type codes'
    0: '',
    1: '°C',
    2: '°F',
    3: 'K',
    4: 'V',
    5: 'A',
    6: 'W',
    7: 'J',
    8: 'C',
    9: 'VA',
    10: 'nt',
    11: 'lm',
    12: 'lx',
    13: 'cd',
    14: 'kPa',
    15: 'PSI',
    16: 'N',
    17: 'CFM',
    18: 'RPM',
    19: 'Hz',
    20: 'μs',
    21: 'ms',
    22: 's',
    23: 'min',
    24: 'hr',
    25: 'd',
    26: 'week(s)',
    27: 'mil',
    28: 'inches',
    29: 'ft',
    30: 'cu in',
    31: 'cu feet',
    32: 'mm',
    33: 'cm',
    34: 'm',
    35: 'cu cm',
    36: 'cu m',
    37: 'L',
    38: 'fl. oz.',
    39: 'radians',
    40: 'steradians',
    41: 'revolutions',
    42: 'cycles',
    43: 'g',
    44: 'ounce',
    45: 'lb',
    46: 'ft-lb',
    47: 'oz-in',
    48: 'gauss',
    49: 'gilberts',
    50: 'henry',
    51: 'millihenry',
    52: 'farad',
    53: 'microfarad',
    54: 'ohms',
    55: 'siemens',
    56: 'mole',
    57: 'becquerel',
    58: 'ppm',
    60: 'dB',
    61: 'dBA',
    62: 'dBC',
    63: 'Gy',
    64: 'sievert',
    65: 'color temp deg K',
    66: 'bit',
    67: 'kb',
    68: 'mb',
    69: 'gb',
    70: 'byte',
    71: 'kB',
    72: 'mB',
    73: 'gB',
    74: 'word',
    75: 'dword',
    76: 'qword',
    77: 'line',
    78: 'hit',
    79: 'miss',
    80: 'retry',
    81: 'reset',
    82: 'overrun/overflow',
    83: 'underrun',
    84: 'collision',
    85: 'packets',
    86: 'messages',
    87: 'characters',
    88: 'error',
    89: 'uncorrectable error',
    90: 'correctable error',
    91: 'fatal error',
    92: 'grams',
}

sensor_rates = {
    0: '',
    1: ' per us',
    2: ' per ms',
    3: ' per s',
    4: ' per minute',
    5: ' per hour',
    6: ' per day',
}


class SensorReading(object):
    """Representation of the state of a sensor.

    It is initialized by pyghmi internally, it does not make sense for
    a developer to create one of these objects directly.

    It provides the following properties:
    name: UTF-8 string describing the sensor
    units: UTF-8 string describing the units of the sensor (if numeric)
    value: Value of the sensor if numeric
    imprecision: The amount by which the actual measured value may deviate from
        'value' due to limitations in the resolution of the given sensor.
    """

    def __init__(self, reading, suffix):
        self.broken_sensor_ids = {}
        self.health = const.Health.Ok
        self.type = reading['type']
        self.value = None
        self.imprecision = None
        self.states = []
        self.state_ids = []
        self.unavailable = 0
        try:
            self.health = reading['health']
            self.states = reading['states']
            self.state_ids = reading['state_ids']
            self.value = reading['value']
            self.imprecision = reading['imprecision']
        except KeyError:
            pass
        if 'unavailable' in reading:
            self.unavailable = 1
        self.units = suffix
        self.name = reading['name']

    def __repr__(self):
        return repr({
            'value': self.value,
            'states': self.states,
            'state_ids': self.state_ids,
            'units': self.units,
            'imprecision': self.imprecision,
            'name': self.name,
            'type': self.type,
            'unavailable': self.unavailable,
            'health': self.health
        })

    def simplestring(self):
        """Return a summary string of the reading.

        This is intended as a sampling of how the data could be presented by
        a UI.  It's intended to help a developer understand the relation
        between the attributes of a sensor reading if it is not quite clear
        """
        repr = self.name + ": "
        if self.value is not None:
            repr += str(self.value)
            repr += " ± " + str(self.imprecision)
            repr += self.units
        for state in self.states:
            repr += state + ","
        if self.health >= const.Health.Failed:
            repr += '(Failed)'
        elif self.health >= const.Health.Critical:
            repr += '(Critical)'
        elif self.health >= const.Health.Warning:
            repr += '(Warning)'
        return repr


class SDREntry(object):
    """Represent a single entry in the IPMI SDR.

    This is created and consumed by pyghmi internally, there is no reason for
    external code to pay attention to this class.
    """

    def __init__(self, entrybytes, ipmicmd, reportunsupported=False):
        # ignore record id for now, we only care about the sensor number for
        # moment
        self.reportunsupported = reportunsupported
        self.ipmicmd = ipmicmd
        if entrybytes[2] != 0x51:
            # only recognize '1.5', the only version defined at time of writing
            raise NotImplementedError
        self.rectype = entrybytes[3]
        self.linearization = None
        # most important to get going are 1, 2, and 11
        self.sdrtype = TYPE_SENSOR  # assume a sensor
        if self.rectype == 1:  # full sdr
            self.full_decode(entrybytes[5:])
        elif self.rectype == 2:  # full sdr
            self.compact_decode(entrybytes[5:])
        elif self.rectype == 8:  # entity association
            self.association_decode(entrybytes[5:])
        elif self.rectype == 0x11:  # FRU locator
            self.fru_decode(entrybytes[5:])
        elif self.rectype == 0x12:  # Management controller
            self.mclocate_decode(entrybytes[5:])
        elif self.rectype == 0xc0:  # OEM format
            self.sdrtype = TYPE_UNKNOWN   # assume undefined
            self.oem_decode(entrybytes[5:])
        elif self.reportunsupported:
            raise NotImplementedError
        else:
            self.sdrtype = TYPE_UNKNOWN

    @property
    def name(self):
        if self.sdrtype == TYPE_SENSOR:
            return self.sensor_name
        elif self.sdrtype == TYPE_FRU:
            return self.fru_name
        else:
            return "UNKNOWN"

    def oem_decode(self, entry):
        mfgid = entry[0] + (entry[1] << 8) + (entry[2] << 16)
        if self.reportunsupported:
            raise NotImplementedError("No support for mfgid %X" % mfgid)

    def mclocate_decode(self, entry):
        # For now, we don't have use for MC locator records
        # we'll ignore them at the moment
        self.sdrtype = TYPE_UNKNOWN
        pass

    def fru_decode(self, entry):
        # table 43-7 FRU Device Locator
        self.sdrtype = TYPE_FRU
        self.fru_name = self.tlv_decode(entry[10], entry[11:])
        self.fru_number = entry[1]
        self.fru_logical = (entry[2] & 0b10000000) == 0b10000000
        # 0x8  to 0x10..  0 unspecified except on 0x10, 1 is dimm
        self.fru_type_and_modifier = (entry[5] << 8) + entry[6]

    def association_decode(self, entry):
        # table 43-4 Entity Associaition Record
        # TODO(jbjohnso): actually represent this data
        self.sdrtype = TYPE_UNKNOWN

    def compact_decode(self, entry):
        # table 43-2 compact sensor record
        self._common_decode(entry)
        self.sensor_name = self.tlv_decode(entry[26], entry[27:])

    def assert_trap_value(self, offset):
        trapval = (self.sensor_type_number << 16) + (self.reading_type << 8)
        return trapval + offset

    def _common_decode(self, entry):
        # compact and full are very similar
        # this function handles the common aspects of compact and full
        # offsets from spec, minus 6
        self.sensor_number = entry[2]
        self.entity = ipmiconst.entity_ids.get(
            entry[3], 'Unknown entity {0}'.format(entry[3]))
        self.sensor_type_number = entry[7]
        try:
            self.sensor_type = ipmiconst.sensor_type_codes[entry[7]]
        except KeyError:
            self.sensor_type = "UNKNOWN type " + str(entry[7])
        self.reading_type = entry[8]  # table 42-1
        # 0: unspecified
        # 1: generic threshold based
        # 0x6f: discrete sensor-specific from table 42-3, sensor offsets
        # all others per table 42-2, generic discrete
        # numeric format is one of:
        # 0 - unsigned, 1 - 1s complement, 2 - 2s complement, 3 - ignore number
        # compact records are supposed to always write it as '3', presumably
        # to allow for the concept of a compact record with a numeric format
        # even though numerics are not allowed today.  Some implementations
        # violate the spec and do something other than 3 today.  Tolerate
        # the violation under the assumption that things are not so hard up
        # that there will ever be a need for compact sensors supporting numeric
        # values
        if self.rectype == 2:
            self.numeric_format = 3
        else:
            self.numeric_format = (entry[15] & 0b11000000) >> 6
        self.sensor_rate = sensor_rates[(entry[15] & 0b111000) >> 3]
        self.unit_mod = ""
        if (entry[15] & 0b110) == 0b10:  # unit1 by unit2
            self.unit_mod = "/"
        elif (entry[15] & 0b110) == 0b100:
            # combine the units by multiplying, SI nomenclature is either spac
            # or hyphen, so go with space
            self.unit_mod = " "
        self.percent = ''
        if entry[15] & 1 == 1:
            self.percent = '% '
        self.baseunit = unit_types[entry[16]]
        self.modunit = unit_types[entry[17]]
        self.unit_suffix = self.percent + self.baseunit + self.unit_mod + \
            self.modunit

    def full_decode(self, entry):
        # offsets are table from spec, minus 6
        # TODO(jbjohnso): table 43-13, put in constants to interpret entry[3]
        self._common_decode(entry)
        # now must extract the formula data to transform values
        # entry[18 to entry[24].
        # if not linear, must use get sensor reading factors
        # TODO(jbjohnso): the various other values
        self.sensor_name = self.tlv_decode(entry[42], entry[43:])
        self.linearization = entry[18] & 0b1111111
        if self.linearization <= 11:
            # the enumuration of linear sensors goes to 11,
            # static formula parameters are applicable, decode them
            # if 0x70, then the sesor reading will have to get the
            # factors on the fly.
            # the formula could apply if we bother with nominal
            # reading interpretation
            self.decode_formula(entry[19:25])

    def _decode_state(self, state):
        mapping = ipmiconst.generic_type_offsets
        try:
            if self.reading_type in mapping:
                desc = mapping[self.reading_type][state]['desc']
                health = mapping[self.reading_type][state]['severity']
            elif self.reading_type == 0x6f:
                mapping = ipmiconst.sensor_type_offsets
                desc = mapping[self.sensor_type_number][state]['desc']
                health = mapping[self.sensor_type_number][state]['severity']
            else:
                desc = "Unknown state %d" % state
                health = const.Health.Warning
        except KeyError:
            desc = "Unknown state %d for reading type %d/sensor type %d" % (
                state, self.reading_type, self.sensor_type_number)
            health = const.Health.Warning
        return desc, health

    def decode_sensor_reading(self, reading):
        numeric = None
        output = {
            'name': self.sensor_name,
            'type': self.sensor_type,
            'id': self.sensor_number,
        }
        if reading[1] & 0b100000:
            output['unavailable'] = 1
            return SensorReading(output, self.unit_suffix)
        if self.numeric_format == 2:
            numeric = twos_complement(reading[0], 8)
        elif self.numeric_format == 1:
            numeric = ones_complement(reading[0], 8)
        elif self.numeric_format == 0:
            numeric = reading[0]
        discrete = True
        if numeric is not None:
            lowerbound = numeric - (0.5 + (self.tolerance / 2.0))
            upperbound = numeric + (0.5 + (self.tolerance / 2.0))
            lowerbound = self.decode_value(lowerbound)
            upperbound = self.decode_value(upperbound)
            output['value'] = (lowerbound + upperbound) / 2.0
            output['imprecision'] = output['value'] - lowerbound
            discrete = False
        upper = 'upper'
        lower = 'lower'
        if self.linearization == 7:
            # if the formula is 1/x, then the intuitive sense of upper and
            # lower are backwards
            upper = 'lower'
            lower = 'upper'
        output['states'] = []
        output['state_ids'] = []
        output['health'] = const.Health.Ok
        if discrete:
            for state in range(8):
                if reading[2] & (0b1 << state):
                    statedesc, health = self._decode_state(state)
                    output['health'] |= health
                    output['states'].append(statedesc)
                    output['state_ids'].append(self.assert_trap_value(state))
            if len(reading) > 3:
                for state in range(7):
                    if reading[3] & (0b1 << state):
                        statedesc, health = self._decode_state(state + 8)
                        output['health'] |= health
                        output['states'].append(statedesc)
                        output['state_ids'].append(
                            self.assert_trap_value(state + 8))
        else:
            if reading[2] & 0b1:
                output['health'] |= const.Health.Warning
                output['states'].append(lower + " non-critical threshold")
                output['state_ids'].append(self.assert_trap_value(1))
            if reading[2] & 0b10:
                output['health'] |= const.Health.Critical
                output['states'].append(lower + " critical threshold")
                output['state_ids'].append(self.assert_trap_value(2))
            if reading[2] & 0b100:
                output['health'] |= const.Health.Failed
                output['states'].append(lower + " non-recoverable threshold")
                output['state_ids'].append(self.assert_trap_value(3))
            if reading[2] & 0b1000:
                output['health'] |= const.Health.Warning
                output['states'].append(upper + " non-critical threshold")
                output['state_ids'].append(self.assert_trap_value(4))
            if reading[2] & 0b10000:
                output['health'] |= const.Health.Critical
                output['states'].append(upper + " critical threshold")
                output['state_ids'].append(self.assert_trap_value(5))
            if reading[2] & 0b100000:
                output['health'] |= const.Health.Failed
                output['states'].append(upper + " non-recoverable threshold")
                output['state_ids'].append(self.assert_trap_value(6))
        return SensorReading(output, self.unit_suffix)

    def _set_tmp_formula(self, value):
        rsp = self.ipmicmd.raw_command(netfn=4, command=0x23,
                                       data=(self.sensor_number, value))
        # skip next reading field, not used in on-demand situation
        self.decode_formula(rsp['data'][1:])

    def decode_value(self, value):
        # Take the input value and return meaningful value
        linearization = self.linearization
        if linearization > 11:  # direct calling code to get factors
            # for now, we will get the factors on demand
            # the facility is engineered such that at construction
            # time the entire BMC table should be fetchable in a reasonable
            # fashion.  However for now opt for retrieving rows as needed
            # rather than tracking all that information for a relatively
            # rare behavior
            self._set_tmp_formula(value)
            linearization = 0
        # time to compute the pre-linearization value.
        decoded = float((value * self.m + self.b) *
                        (10 ** self.resultexponent))
        if linearization == 0:
            return decoded
        elif linearization == 1:
            return math.log(decoded)
        elif linearization == 2:
            return math.log(decoded, 10)
        elif linearization == 3:
            return math.log(decoded, 2)
        elif linearization == 4:
            return math.exp(decoded)
        elif linearization == 5:
            return 10 ** decoded
        elif linearization == 6:
            return 2 ** decoded
        elif linearization == 7:
            return 1 / decoded
        elif linearization == 8:
            return decoded ** 2
        elif linearization == 9:
            return decoded ** 3
        elif linearization == 10:
            return math.sqrt(decoded)
        elif linearization == 11:
            return decoded ** (1.0/3)
        else:
            raise NotImplementedError

    def decode_formula(self, entry):
        self.m = \
            twos_complement(entry[0] + ((entry[1] & 0b11000000) << 2), 10)
        self.tolerance = entry[1] & 0b111111
        self.b = \
            twos_complement(entry[2] + ((entry[3] & 0b11000000) << 2), 10)
        self.accuracy = (entry[3] & 0b111111) + \
            (entry[4] & 0b11110000) << 2
        self.accuracyexp = (entry[4] & 0b1100) >> 2
        self.direction = entry[4] & 0b11
        # 0 = n/a, 1 = input, 2 = output
        self.resultexponent = twos_complement((entry[5] & 0b11110000) >> 4, 4)
        bexponent = twos_complement(entry[5] & 0b1111, 4)
        # might as well do the math to 'b' now rather than wait for later
        self.b = self.b * (10**bexponent)

    def tlv_decode(self, tlv, data):
        # Per IPMI 'type/length byte format
        ipmitype = (tlv & 0b11000000) >> 6
        if not len(data):
            return ""
        if ipmitype == 0:  # Unicode per 43.15 in ipmi 2.0 spec
            # the spec is not specific about encoding, assuming utf8
            return unicode(struct.pack("%dB" % len(data), *data), "utf_8")
        elif ipmitype == 1:  # BCD '+'
            tmpl = "%02X" * len(data)
            tstr = tmpl % tuple(data)
            tstr = tstr.replace("A", " ").replace("B", "-").replace("C", ".")
            return tstr.replace("D", ":").replace("E", ",").replace("F", "_")
        elif ipmitype == 2:  # 6 bit ascii, start at 0x20
            # the ordering is very peculiar and is best understood from
            # IPMI SPEC "6-bit packed ascii example
            tstr = ""
            while len(data) >= 3:  # the packing only works with 3 byte chunks
                tstr += chr((data[0] & 0b111111) + 0x20)
                tstr += chr(((data[1] & 0b1111) << 2) +
                            (data[0] >> 6) + 0x20)
                tstr += chr(((data[2] & 0b11) << 4) +
                            (data[1] >> 4) + 0x20)
                tstr += chr((data[2] >> 2) + 0x20)
            return tstr
        elif ipmitype == 3:  # ACSII+LATIN1
            return struct.pack("%dB" % len(data), *data)


class SDR(object):
    """Examine the state of sensors managed by a BMC

    Presents the data from sensor read commands as directed by the SDR in a
    reasonable format.  This module is used by the command module, and is not
    intended for consumption by external code directly

    :param ipmicmd: A Command class object
    """
    def __init__(self, ipmicmd):
        self.ipmicmd = weakref.proxy(ipmicmd)
        self.sensors = {}
        self.fru = {}
        self.read_info()

    def read_info(self):
        # first, we want to know the device id
        rsp = self.ipmicmd.xraw_command(netfn=6, command=1)
        rsp['data'] = bytearray(rsp['data'])
        self.device_id = rsp['data'][0]
        self.device_rev = rsp['data'][1] & 0b111
        # Going to ignore device available until get sdr command
        # since that provides usefully distinct state and this does not
        self.fw_major = rsp['data'][2] & 0b1111111
        self.fw_minor = "%02X" % rsp['data'][3]  # BCD encoding, oddly enough
        self.ipmiversion = rsp['data'][4]  # 51h = 1.5, 02h = 2.0
        self.mfg_id = rsp['data'][8] << 16 + rsp['data'][7] << 8 + \
            rsp['data'][6]
        self.prod_id = rsp['data'][10] << 8 + rsp['data'][9]
        if len(rsp['data']) > 11:
            self.aux_fw = self.decode_aux(rsp['data'][11:15])
        if rsp['data'][1] & 0b10000000 and rsp['data'][5] & 0b10 == 0:
            # The device has device sdrs, also does not support SDR repository
            # device, so we are meant to use an alternative mechanism to get
            # SDR data
            if rsp['data'][5] & 1:
                # The device has sensor device support, so in theory we should
                # be able to proceed
                # However at the moment, we haven't done so
                raise NotImplementedError
            return
            # We have Device SDR, without SDR Repository device, but
            # also without sensor device support, no idea how to
            # continue
        self.get_sdr()

    def get_sdr_reservation(self):
        rsp = self.ipmicmd.raw_command(netfn=0x0a, command=0x22)
        if rsp['code'] != 0:
            raise exc.IpmiException(rsp['error'])
        return rsp['data'][0] + (rsp['data'][1] << 8)

    def get_sdr(self):
        repinfo = self.ipmicmd.xraw_command(netfn=0x0a, command=0x20)
        repinfo['data'] = bytearray(repinfo['data'])
        if (repinfo['data'][0] != 0x51):
            # we only understand SDR version 51h, the only version defined
            # at time of this writing
            raise NotImplementedError
        # NOTE(jbjohnso): we actually don't need to care about 'numrecords'
        # since FFFF marks the end explicitly
        # numrecords = (rsp['data'][2] << 8) + rsp['data'][1]
        # NOTE(jbjohnso): don't care about 'free space' at the moment
        # NOTE(jbjohnso): most recent timstamp data for add and erase could be
        # handy to detect cache staleness, but for now will assume invariant
        # over life of session
        # NOTE(jbjohnso): not looking to support the various options in op
        # support, ignore those for now, reservation if some BMCs can't read
        # full SDR in one slurp
        recid = 0
        rsvid = 0  # partial 'get sdr' will require this
        offset = 0
        size = 0xff
        chunksize = 128
        self.broken_sensor_ids = {}
        while recid != 0xffff:  # per 33.12 Get SDR command, 0xffff marks end
            newrecid = 0
            currlen = 0
            sdrdata = bytearray()
            while True:  # loop until SDR fetched wholly
                if size != 0xff and rsvid == 0:
                    rsvid = self.get_sdr_reservation()
                rqdata = [rsvid & 0xff, rsvid >> 8,
                          recid & 0xff, recid >> 8,
                          offset, size]
                sdrrec = self.ipmicmd.raw_command(netfn=0x0a, command=0x23,
                                                  data=rqdata)
                if sdrrec['code'] == 0xca:
                    if size == 0xff:  # get just 5 to get header to know length
                        size = 5
                    elif size > 5:
                        size /= 2
                        # push things over such that it's less
                        # likely to be just 1 short of a read
                        # and incur a whole new request
                        size += 2
                        chunksize = size
                    continue
                if sdrrec['code'] == 0xc5:  # need a new reservation id
                    rsvid = 0
                    continue
                if sdrrec['code'] != 0:
                    raise exc.IpmiException(sdrrec['error'])
                if newrecid == 0:
                    newrecid = (sdrrec['data'][1] << 8) + sdrrec['data'][0]
                if currlen == 0:
                    currlen = sdrrec['data'][6] + 5  # compensate for header
                sdrdata.extend(sdrrec['data'][2:])
                # determine next offset to use based on current offset and the
                # size used last time.
                offset += size
                if offset >= currlen:
                    break
                if size == 5 and offset == 5:
                    # bump up size after header retrieval
                    size = chunksize
                if (offset + size) > currlen:
                    size = currlen - offset
            self.add_sdr(sdrdata)
            offset = 0
            if size != 0xff:
                size = 5
            if newrecid == recid:
                raise exc.BmcErrorException("Incorrect SDR record id from BMC")
            recid = newrecid
        for sid in self.broken_sensor_ids:
            try:
                del self.sensors[sid]
            except KeyError:
                pass

    def get_sensor_numbers(self):
        return self.sensors.iterkeys()

    def add_sdr(self, sdrbytes):
        newent = SDREntry(sdrbytes, self.ipmicmd)
        if newent.sdrtype == TYPE_SENSOR:
            id = newent.sensor_number
            if id in self.sensors:
                self.broken_sensor_ids[id] = True
                return
            self.sensors[id] = newent
        elif newent.sdrtype == TYPE_FRU:
            id = newent.fru_number
            if id in self.fru:
                self.broken_sensor_ids[id] = True
                return
            self.fru[id] = newent

    def decode_aux(self, auxdata):
        # This is where manufacturers can add their own
        # decode information
        return "".join(hex(x) for x in auxdata)

if __name__ == "__main__":  # test code
    import os
    import pyghmi.ipmi.command as ipmicmd
    import sys
    password = os.environ['IPMIPASSWORD']
    bmc = sys.argv[1]
    user = sys.argv[2]
    ipmicmd = ipmicmd.Command(bmc=bmc, userid=user, password=password)
    sdr = SDR(ipmicmd)
    for number in sdr.get_sensor_numbers():
        rsp = ipmicmd.raw_command(command=0x2d, netfn=4, data=(number,))
        if 'error' in rsp:
            continue
        reading = sdr.sensors[number].decode_sensor_reading(rsp['data'])
        if reading is not None:
            print(repr(reading))



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyghmi.ipmi.oem.lenovo.inventory import EntryField, \
    parse_inventory_category_entry

psu_type = {
    0b0001: "Other",
    0b0010: "Unknown",
    0b0011: "Linear",
    0b0100: "Switching",
    0b0101: "Battery",
    0b0110: "UPS",
    0b0111: "Converter",
    0b1000: "Regulator",
}
psu_status = {
    0b001: "Other",
    0b010: "Unknown",
    0b011: "OK",
    0b100: "Non-critical",
    0b101: "Critical; power supply has failed and has been taken off-line"
}
psu_voltage_range_switch = {
    0b0001: "Other",
    0b0010: "Unknown",
    0b0011: "Manual",
    0b0100: "Auto-switch",
    0b0101: "Wide range",
    0b0110: "Not applicable"
}


def psu_status_word_slice(w, s, e):
    return int(w[-e-1:-s], 2)


def psu_status_word_bit(w, b):
    return int(w[-b-1])


def psu_status_word_parser(word):
    fields = {}
    word = "{0:016b}".format(word)

    fields["DMTF Power Supply Type"] = \
        psu_type.get(psu_status_word_slice(word, 10, 13), "Invalid")

    # fields["Status"] = \
    #    psu_status.get(psu_status_word_slice(word, 7, 9), "Invalid")

    fields["DMTF Input Voltage Range"] = \
        psu_voltage_range_switch.get(
            psu_status_word_slice(word, 3, 6),
            "Invalid"
        )

    # Power supply is unplugged from the wall
    fields["Unplugged"] = \
        bool(psu_status_word_bit(word, 2))

    # fields["Power supply is present"] = \
    #    bool(psu_status_word_bit(word, 1))

    # Power supply is hot-replaceable
    fields["Hot Replaceable"] = \
        bool(psu_status_word_bit(word, 0))

    return fields

psu_fields = (
    EntryField("index", "B"),
    EntryField("Presence State", "B", presence=True),
    EntryField("Capacity W", "<H"),
    EntryField("Board manufacturer", "18s"),
    EntryField("Board model", "18s"),
    EntryField("Board manufacture date", "10s"),
    EntryField("Board serial number", "34s"),
    EntryField("Board manufacturer revision", "5s"),
    EntryField("Board product name", "10s"),
    EntryField("PSU Asset Tag", "10s"),
    EntryField(
        "PSU Redundancy Status",
        "B",
        valuefunc=lambda v: "Not redundant" if v == 0x00 else "Redundant"
    ),
    EntryField(
        "PSU Status Word",
        "<H",
        valuefunc=psu_status_word_parser, multivaluefunc=True
    )
)


def parse_psu_info(raw):
    return parse_inventory_category_entry(raw, psu_fields)


def get_categories():
    return {
        "psu": {
            "idstr": "Power Supply {0}",
            "parser": parse_psu_info,
            "command": {
                "netfn": 0x06,
                "command": 0x59,
                "data": (0x00, 0xc6, 0x00, 0x00)
            }
        }
    }



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyghmi.ipmi.oem.lenovo.inventory import EntryField, \
    parse_inventory_category_entry

raid_controller_fields = (
    EntryField("ControllerID", "I"),
    EntryField("AdapterType", "B", mapper={
        0x00: "Unknown",
        0x01: "RAIDController"
    }),
    EntryField("SupercapPresence", "B", mapper={
        0x00: "Absent",
        0x01: "Present"
    }),
    EntryField("FlashComponent1Name", "16s"),
    EntryField("FlashComponent1Version", "64s"),
    EntryField("FlashComponent2Name", "16s"),
    EntryField("FlashComponent2Version", "64s"),
    EntryField("FlashComponent3Name", "16s"),
    EntryField("FlashComponent3Version", "64s"),
    EntryField("FlashComponent4Name", "16s"),
    EntryField("FlashComponent4Version", "64s"),
    EntryField("FlashComponent5Name", "16s"),
    EntryField("FlashComponent5Version", "64s"),
    EntryField("FlashComponent6Name", "16s"),
    EntryField("FlashComponent6Version", "64s"),
    EntryField("FlashComponent7Name", "16s"),
    EntryField("FlashComponent7Version", "64s"),
    EntryField("FlashComponent8Name", "16s"),
    EntryField("FlashComponent8Version", "64s")
)


def parse_raid_controller_info(raw):
    return parse_inventory_category_entry(raw, raid_controller_fields)


def get_categories():
    return {
        "raid_controller": {
            "idstr": "RAID Controller {0}",
            "parser": parse_raid_controller_info,
            "countable": False,
            "command": {
                "netfn": 0x06,
                "command": 0x59,
                "data": (0x00, 0xc4, 0x00, 0x00)
            }
        }
    }



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pyghmi.ipmi.oem.lenovo.inventory import EntryField, \
    parse_inventory_category_entry

dimm_fields = (
    EntryField("index", "B"),
    EntryField("manufacture_location", "B"),
    EntryField("channel_number", "B"),
    EntryField("module_type", "10s"),
    EntryField("ddr_voltage", "10s"),
    EntryField("speed", "<h",
               valuefunc=lambda v: str(v) + " MHz"),
    EntryField("capacity_mb", "<h",
               valuefunc=lambda v: v*1024),
    EntryField("manufacturer", "30s"),
    EntryField("serial", ">I",
               valuefunc=lambda v: hex(v)[2:]),
    EntryField("model", "21s"),
    EntryField("reserved", "h", include=False)
)


def parse_dimm_info(raw):
    return parse_inventory_category_entry(raw, dimm_fields)


def get_categories():
    return {
        "dimm": {
            "idstr": "DIMM {0}",
            "parser": parse_dimm_info,
            "command": {
                "netfn": 0x06,
                "command": 0x59,
                "data": (0x00, 0xc1, 0x02, 0x00)
            },
            "workaround_bmc_bug": True
        }
    }



# Copyright 2015 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This provides ability to do HTTPS in a manner like ssh host keys for the
# sake of typical internal management devices.  Compatibility back to python
# 2.6 as is found in commonly used enterprise linux distributions.

import json
import pyghmi.exceptions as pygexc
import socket
import ssl

try:
    import Cookie
    import httplib
except ImportError:
    import http.client as httplib
    import http.cookies as Cookie

__author__ = 'jjohnson2'


class SecureHTTPConnection(httplib.HTTPConnection, object):
    default_port = httplib.HTTPS_PORT

    def __init__(self, host, port=None, key_file=None, cert_file=None,
                 ca_certs=None, strict=None, verifycallback=None, **kwargs):
        if 'timeout' not in kwargs:
            kwargs['timeout'] = 60
        httplib.HTTPConnection.__init__(self, host, port, strict, **kwargs)
        self.cert_reqs = ssl.CERT_NONE  # verification will be done ssh style..
        self._certverify = verifycallback
        self.cookies = {}
        self.stdheaders = {}

    def set_header(self, key, value):
        self.stdheaders[key] = value

    def connect(self):
        plainsock = socket.create_connection((self.host, self.port), 60)
        self.sock = ssl.wrap_socket(plainsock, cert_reqs=self.cert_reqs)
        # txtcert = self.sock.getpeercert()  # currently not possible
        bincert = self.sock.getpeercert(binary_form=True)
        if not self._certverify(bincert):
            raise pygexc.UnrecognizedCertificate('Unknown certificate',
                                                 bincert)

    def getresponse(self):
        rsp = super(SecureHTTPConnection, self).getresponse()
        for hdr in rsp.msg.headers:
            if hdr.startswith('Set-Cookie:'):
                c = Cookie.BaseCookie(hdr[11:])
                for k in c:
                    self.cookies[k] = c[k].value
        return rsp

    def grab_json_response(self, url, data=None):
        if data:
            self.request('POST', url, data)
        else:
            self.request('GET', url)
        rsp = self.getresponse()
        if rsp.status == 200:
            return json.loads(rsp.read())
        rsp.read()
        return {}

    def request(self, method, url, body=None, headers=None):
        if headers is None:
            headers = self.stdheaders.copy()
        if method == 'GET' and 'Content-Type' in headers:
            del headers['Content-Type']
        if self.cookies:
            cookies = []
            for ckey in self.cookies:
                cookies.append('{0}={1}'.format(ckey, self.cookies[ckey]))
            headers['Cookie'] = '; '.join(cookies)
        return super(SecureHTTPConnection, self).request(method, url, body,
                                                         headers)



# vim: tabstop=4 shiftwidth=4 softtabstop=4

# Copyright 2017 Lenovo
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ctypes import addressof, c_int, c_long, c_short, c_ubyte, c_uint
from ctypes import cast, create_string_buffer, POINTER, pointer, sizeof
from ctypes import Structure
import fcntl
import pyghmi.ipmi.private.util as iutil
from select import select


class IpmiMsg(Structure):
    _fields_ = [('netfn', c_ubyte),
                ('cmd', c_ubyte),
                ('data_len', c_short),
                ('data', POINTER(c_ubyte))]


class IpmiSystemInterfaceAddr(Structure):
    _fields_ = [('addr_type', c_int),
                ('channel', c_short),
                ('lun', c_ubyte)]


class IpmiRecv(Structure):
    _fields_ = [('recv_type', c_int),
                ('addr', POINTER(IpmiSystemInterfaceAddr)),
                ('addr_len', c_uint),
                ('msgid', c_long),
                ('msg', IpmiMsg)]


class IpmiReq(Structure):
    _fields_ = [('addr', POINTER(IpmiSystemInterfaceAddr)),
                ('addr_len', c_uint),
                ('msgid', c_long),
                ('msg', IpmiMsg)]


_IONONE = 0
_IOWRITE = 1
_IOREAD = 2
IPMICTL_SET_MY_ADDRESS_CMD = _IOREAD << 30 | sizeof(c_uint) << 16 | \
    ord('i') << 8 | 17  # from ipmi.h
IPMICTL_SEND_COMMAND = _IOREAD << 30 | sizeof(IpmiReq) << 16 | \
    ord('i') << 8 | 13  # from ipmi.h
# next is really IPMICTL_RECEIVE_MSG_TRUNC, but will only use that
IPMICTL_RECV = (_IOWRITE | _IOREAD) << 30 | sizeof(IpmiRecv) << 16 | \
    ord('i') << 8 | 11  # from ipmi.h
BMC_SLAVE_ADDR = c_uint(0x20)
CURRCHAN = 0xf
ADDRTYPE = 0xc


class Session(object):
    def __init__(self, devnode='/dev/ipmi0'):
        """Create a local session inband

        :param: devnode: The path to the ipmi device
        """
        self.ipmidev = open(devnode, 'rw')
        fcntl.ioctl(self.ipmidev, IPMICTL_SET_MY_ADDRESS_CMD, BMC_SLAVE_ADDR)
        # the interface is initted, create some reusable memory for our session
        self.databuffer = create_string_buffer(4096)
        self.req = IpmiReq()
        self.rsp = IpmiRecv()
        self.addr = IpmiSystemInterfaceAddr()
        self.req.msg.data = cast(addressof(self.databuffer), POINTER(c_ubyte))
        self.rsp.msg.data = self.req.msg.data
        self.userid = None
        self.password = None

    def await_reply(self):
        rd, _, _ = select((self.ipmidev,), (), (), 1)
        while not rd:
            rd, _, _ = select((self.ipmidev,), (), (), 1)

    @property
    def parsed_rsp(self):
        response = {'netfn': self.rsp.msg.netfn, 'command': self.rsp.msg.cmd,
                    'code': ord(self.databuffer.raw[0]),
                    'data': list(bytearray(
                        self.databuffer.raw[1:self.rsp.msg.data_len]))}
        errorstr = iutil.get_ipmi_error(response)
        if errorstr:
            response['error'] = errorstr
        return response

    def raw_command(self,
                    netfn,
                    command,
                    data=(),
                    bridge_request=None,
                    retry=True,
                    delay_xmit=None,
                    timeout=None,
                    waitall=False):
        self.addr.channel = CURRCHAN
        self.addr.addr_type = ADDRTYPE
        self.req.addr_len = sizeof(IpmiSystemInterfaceAddr)
        self.req.addr = pointer(self.addr)
        self.req.msg.netfn = netfn
        self.req.msg.cmd = command
        data = buffer(bytearray(data))
        self.databuffer[:len(data)] = data[:len(data)]
        self.req.msg.data_len = len(data)
        fcntl.ioctl(self.ipmidev, IPMICTL_SEND_COMMAND, self.req)
        self.await_reply()
        self.rsp.msg.data_len = 4096
        self.rsp.addr = pointer(self.addr)
        self.rsp.addr_len = sizeof(IpmiSystemInterfaceAddr)
        fcntl.ioctl(self.ipmidev, IPMICTL_RECV, self.rsp)
        return self.parsed_rsp


def main():
    a = Session('/dev/ipmi0')
    print(repr(a.raw_command(0, 1)))


if __name__ == '__main__':
    main()
